/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.expression.converter.rdb;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.expression.api.BinaryOperator;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.UnaryOperator;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.Concat;
import org.objectweb.medor.expression.lib.FirstLocate;
import org.objectweb.medor.expression.lib.IndexedLocate;
import org.objectweb.medor.expression.lib.Length;
import org.objectweb.medor.expression.lib.Like;
import org.objectweb.medor.expression.lib.Mod;
import org.objectweb.medor.expression.lib.Substring;
import org.objectweb.medor.expression.lib.StringUpper;
import org.objectweb.medor.expression.lib.StringLower;
import org.objectweb.medor.expression.converter.api.Expression2WhereClause;

/**
 * This class implements the Expression2WhereClause interface.
 *
 * @author A. Lefebvre
 */
public class Expression2WhereClauseImpl implements Expression2WhereClause {

	private final static String LAND = "&";
	private final static String LCOND_AND = "&&";
	private final static String AND = "AND";
	private final static String EQUAL = "=";
	private final static String LEQUAL = "==";
	private final static String LOR = "|";
	private final static String LCOND_OR = "||";
	private final static String OR = "OR";
	private final static String NOT = "NOT";

	public final String convertExp2WhereClause(Expression exp,
											   RdbAdapter rdbAdapter)
		throws ExpressionException {
		StringBuffer sb = new StringBuffer();
		convertExp2WhereClause(exp, rdbAdapter, sb);
		return sb.toString();
	}

	/**
	 * These method build a where clause in a StrinbBuffer from a medor
	 * expression.
	 * @param exp is the medor expression to translate
	 * @param rdbAdapter is the RdbAdapter permetting to
	 * @param sb
	 * @throws ExpressionException
	 */
	public void convertExp2WhereClause(Expression exp,
									   RdbAdapter rdbAdapter,
									   StringBuffer sb)
		throws ExpressionException {
		if (exp instanceof ParameterOperand) {
			sb.append("?");
		} else if (exp instanceof BasicOperand) {
			BasicOperand o = (BasicOperand) exp;
			if (rdbAdapter != null) {
				switch (o.getType().getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
					sb.append(rdbAdapter.getValueAsSQLString(o.getBoolean()));
					break;
				case PType.TYPECODE_CHAR:
					sb.append(rdbAdapter.getValueAsSQLString(o.getChar()));
					break;
				case PType.TYPECODE_BYTE:
					sb.append(rdbAdapter.getValueAsSQLString(o.getByte()));
					break;
				case PType.TYPECODE_SHORT:
					sb.append(rdbAdapter.getValueAsSQLString(o.getShort()));
					break;
				case PType.TYPECODE_INT:
					sb.append(rdbAdapter.getValueAsSQLString(o.getInt()));
					break;
				case PType.TYPECODE_LONG:
					sb.append(rdbAdapter.getValueAsSQLString(o.getLong()));
					break;
				case PType.TYPECODE_FLOAT:
					sb.append(rdbAdapter.getValueAsSQLString(o.getFloat()));
					break;
				case PType.TYPECODE_DOUBLE:
					sb.append(rdbAdapter.getValueAsSQLString(o.getDouble()));
					break;
				default:
					sb.append(rdbAdapter.getValueAsSQLString(
						o.getObject(),
						o.getType().getTypeCode()));
				}
			} else {
				switch (o.getType().getTypeCode()) {
				case PType.TYPECODE_OBJBOOLEAN:
				case PType.TYPECODE_BOOLEAN:
					sb.append((o.getBoolean() ? "1" : "0"));
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					sb.append(o.getValueAsString());
					break;
				default:
					sb.append("'" + o.getValueAsString() + "'");
					break;
				}
			}
		} else if (exp instanceof Like) {
			Like l = (Like) exp;
			sb.append("(");
			convertExp2WhereClause(l.getExpression(0), rdbAdapter, sb);
			sb.append(" ");
			sb.append(l.getOperatorString());
			sb.append(" ");
			convertExp2WhereClause(l.getExpression(1), rdbAdapter, sb);
			if (l.getOperandNumber() == 3) {
				sb.append(" ESCAPE ");
				convertExp2WhereClause(l.getExpression(2), rdbAdapter, sb);
			}
			sb.append(")");
		} else if (exp instanceof FirstLocate) {//generic
			Operator op = (Operator) exp;
			if (rdbAdapter == null) {
				sb.append(toSqlOperator(op.getOperatorString()));
				sb.append("(");
				convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
				sb.append(" in ");
				convertExp2WhereClause(op.getExpression(1), rdbAdapter, sb);
				sb.append(")");
			} else {
				try {
					sb.append(
						rdbAdapter.getFirstLocateExpression(
							convertExp2WhereClause(op.getExpression(0), rdbAdapter),
							convertExp2WhereClause(op.getExpression(1), rdbAdapter)
						));
				} catch (RdbAdapterException pe) {
					throw new ExpressionException(pe.getMessage());
				}
			}
		} else if (exp instanceof Concat) {
			Operator op = (Operator) exp;
			if (rdbAdapter == null) {
				sb.append(toSqlOperator(op.getOperatorString()));
				sb.append("(");
				convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
				sb.append(", ");
				convertExp2WhereClause(op.getExpression(1), rdbAdapter, sb);
				sb.append(")");
			} else {
				sb.append(rdbAdapter.getConcatExpression(
					convertExp2WhereClause(op.getExpression(0), rdbAdapter),
					convertExp2WhereClause(op.getExpression(1), rdbAdapter)
				));
			}
		} else if (exp instanceof Mod) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
			sb.append(", ");
			convertExp2WhereClause(op.getExpression(1), rdbAdapter, sb);
			sb.append(")");
		} else if (exp instanceof BinaryOperator) {
			Operator op = (Operator) exp;
			sb.append("(");
			convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
			sb.append(" ");
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append(" ");
			convertExp2WhereClause(op.getExpression(1), rdbAdapter, sb);
			sb.append(")");
		} else if (exp instanceof UnaryOperator) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
			sb.append(")");
		} else if (exp instanceof Substring) {
			Operator op = (Operator) exp;
			if (rdbAdapter == null) {
				sb.append(toSqlOperator(op.getOperatorString()));
			} else {
				sb.append(rdbAdapter.getSubstringName());
			}
			sb.append("(");
			sb.append(convertExp2WhereClause(op.getExpression(0), rdbAdapter));
			sb.append(", ");
			sb.append(convertExp2WhereClause(op.getExpression(1), rdbAdapter));
			sb.append(", ");
			sb.append(convertExp2WhereClause(op.getExpression(2), rdbAdapter));
			sb.append(")");
		} else if (exp instanceof StringUpper || exp instanceof StringLower) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			sb.append(convertExp2WhereClause(op.getExpression(0), rdbAdapter));
			sb.append(")");
		} else if (exp instanceof StringUpper
			|| exp instanceof StringLower || exp instanceof Length) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			sb.append(convertExp2WhereClause(op.getExpression(0), rdbAdapter));
			sb.append(")");
		} else if (exp instanceof IndexedLocate) {
			Operator op = (Operator) exp;
			if (rdbAdapter == null) {
				sb.append(toSqlOperator(op.getOperatorString()));
				sb.append("(");
				convertExp2WhereClause(op.getExpression(0), rdbAdapter, sb);
				sb.append(", ");
				convertExp2WhereClause(op.getExpression(1), rdbAdapter, sb);
				sb.append(", ");
				convertExp2WhereClause(op.getExpression(2), rdbAdapter, sb);
				sb.append(")");
			} else {
				try {
					sb.append(
						rdbAdapter.getIndexedLocateExpression(
							convertExp2WhereClause(op.getExpression(0), rdbAdapter),
							convertExp2WhereClause(op.getExpression(1), rdbAdapter),
							convertExp2WhereClause(op.getExpression(2), rdbAdapter)
						));
				} catch (RdbAdapterException pe) {
					throw new ExpressionException(pe.getMessage());
				}
			}
		} else
			throw new ExpressionException("Umanaged expression element: " + exp);
	}

	/**
	 * Returns the SQL equivalent of the Java operator String.
	 */
	public String toSqlOperator(String in) {
		if (LAND.equals(in) || LCOND_AND.equals(in)) {
			return AND;
		} else if (LOR.equals(in) || LCOND_OR.equals(in)) {
			return OR;
		} else if (in.equals(LEQUAL)) {
			return EQUAL;
		} else if (in.equals(Operator.NOT)) {
			return NOT;
		} else
			return in;
	}

	/**
	 * These method build a where clause in a StrinbBuffer from a medor
	 * expression.
	 * @param exp is the medor expression to translate
	 * @param rdbAdapterVarName is the RdbAdapter permetting to
	 * @param sb
	 * @throws ExpressionException
	 */
	public void convertExp2WhereClauseBuilder(Expression exp,
											  String rdbAdapterVarName,
											  StringBuffer sb)
		throws ExpressionException {
		if (exp instanceof ParameterOperand) {
			sb.append("?");
		} else if (exp instanceof BasicOperand) {
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getValueAsSQLString(");
			BasicOperand o = (BasicOperand) exp;
			switch (o.getType().getTypeCode()) {
			case PType.TYPECODE_BOOLEAN:
			case PType.TYPECODE_OBJBOOLEAN:
				sb.append(o.getBoolean());
				break;
			case PType.TYPECODE_CHAR:
			case PType.TYPECODE_OBJCHAR:
				sb.append("'");
				sb.append(o.getChar());
				sb.append("'");
				break;
			case PType.TYPECODE_BYTE:
			case PType.TYPECODE_OBJBYTE:
				sb.append("(byte) ");
				sb.append(o.getByte());
				break;
			case PType.TYPECODE_OBJSHORT:
				sb.append("(short) ");
				sb.append(o.getShort());
				break;
			case PType.TYPECODE_INT:
			case PType.TYPECODE_OBJINT:
				sb.append(o.getInt());
				break;
			case PType.TYPECODE_LONG:
			case PType.TYPECODE_OBJLONG:
				sb.append(o.getLong());
				sb.append("l");
				break;
			case PType.TYPECODE_FLOAT:
			case PType.TYPECODE_OBJFLOAT:
				sb.append("(float) ");
				sb.append(o.getFloat());
				break;
			case PType.TYPECODE_DOUBLE:
			case PType.TYPECODE_OBJDOUBLE:
				sb.append("(double) ");
				sb.append(o.getDouble());
				break;
			case PType.TYPECODE_STRING:
				sb.append("\"");
				sb.append(o.getString());
				sb.append("\", ");
				sb.append(PTypeSpace.STRING.getProgName());
				break;
			case PType.TYPECODE_DATE:
				sb.append("new java.util.Date(");
				sb.append(o.getDate().getTime());
				sb.append("), ");
				sb.append(PTypeSpace.DATE.getProgName());
				break;
			case PType.TYPECODE_BIGINTEGER:
				sb.append("new java.math.BigInteger(");
				sb.append(o.getBigInteger().toString());
				sb.append("), ");
				sb.append(PTypeSpace.BIGINTEGER.getProgName());
				break;
			case PType.TYPECODE_BIGDECIMAL:
				sb.append("new java.math.BigDecimal(");
				sb.append(o.getBigDecimal().toString());
				sb.append("), ");
				sb.append(PTypeSpace.BIGDECIMAL.getProgName());
				break;
			default:
				sb.append("null");
			}
			sb.append(") + \"");
		} else if (exp instanceof Like) {
			Like l = (Like) exp;
			sb.append("(");
			convertExp2WhereClauseBuilder(l.getExpression(0), rdbAdapterVarName, sb);
			sb.append(l.getOperatorString());
			convertExp2WhereClauseBuilder(l.getExpression(1), rdbAdapterVarName, sb);
			if (l.getOperandNumber() == 3) {
				sb.append(" ESCAPE ");
				convertExp2WhereClauseBuilder(l.getExpression(2), rdbAdapterVarName, sb);
				sb.append(")");
			} else {
				sb.append(")");
			}
		} else if (exp instanceof FirstLocate) {//generic
			Operator op = (Operator) exp;
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getFirstLocateExpression(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(") + \"");
		} else if (exp instanceof Concat) {
			Operator op = (Operator) exp;
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getConcatExpression(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(") + \"");
		} else if (exp instanceof Mod) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof BinaryOperator) {
			Operator op = (Operator) exp;
			sb.append("(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(" ");
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append(" ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof UnaryOperator) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof Substring) {
			Operator op = (Operator) exp;
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getSubstringName() + \"");
			sb.append("(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(2), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof Length) {
			Operator op = (Operator) exp;
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getLengthOperator() + \"");
			sb.append("(");
 			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof StringUpper
			|| exp instanceof StringLower) {
			Operator op = (Operator) exp;
			sb.append(toSqlOperator(op.getOperatorString()));
			sb.append("(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(")");
		} else if (exp instanceof IndexedLocate) {
			Operator op = (Operator) exp;
			sb.append("\" + ");
			sb.append(rdbAdapterVarName);
			sb.append(".getConcatExpression(");
			convertExp2WhereClauseBuilder(op.getExpression(0), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(1), rdbAdapterVarName, sb);
			sb.append(", ");
			convertExp2WhereClauseBuilder(op.getExpression(2), rdbAdapterVarName, sb);
			sb.append(") + \"");
		} else
			throw new ExpressionException("Umanaged expression element: " + exp);
	}
}
