/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2003 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, A. Lefebvre
 */

package org.objectweb.medor.expression.lib;

import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.TypingException;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.type.api.PTypeSpace;

/**
 * Operator representing the search for a substring within a given input string,
 * starting at a given index.
 */
public class IndexedLocate extends BasicOperator {

	/**
	 * Constructs an IndexedLocate operator, representing the search for a
	 * substring in an input string, starting at a given index.
	 * <p>It is the equivalent of the Java indexOf String manipulation method.
	 * <p>If the substring argument occurs as a substring within the
	 * inputString at a starting index no smaller than
	 * <code>fromIndex</code>, then the index of the first character
	 * of the first such substring is returned. If it does not occur
	 * as a substring starting at <code>fromIndex</code> or beyond,
	 * <code>-1</code> is returned.
	 * <p>Unlike in Java, the index starts with 1 and not 0.
	 * @param   inputString the String in which to look.
	 * @param   substring         the substring to search for.
	 * @param   fromIndex   the index to start the search from.

	 */
	public IndexedLocate(Expression inputString, Expression substring,
						 Expression fromIndex) {
		super(PTypeSpace.INT, new Expression[]{inputString, substring, fromIndex});
	}

	public IndexedLocate() {
		super(PTypeSpace.INT);
	}

	public Operand evaluate(ParameterOperand[] pos, Object o)
		throws ExpressionException {
		try {
			result.setValue(
					evaluate(
							getExpression(0).evaluate(pos, o).getString(),
							getExpression(1).evaluate(pos, o).getString(),
							getExpression(2).evaluate(pos, o).getInt()));
		}
		catch (NullPointerException e) {
			// result == null or missing children objects
			throw new ExpressionException("Unevaluable Expression: May not be compiled");
		}
		return result;
	}

	public Operand compileExpression()
		throws ExpressionException, MalformedExpressionException {
		if ((getExpression(0) != null)
				&& (getExpression(1) != null)
				&& (getExpression(2) != null)) {
			getExpression(0).compileExpression();
			getExpression(1).compileExpression();
			getExpression(2).compileExpression();
			if ((getExpression(0).getType().getTypeCode() ==
				PType.TYPECODE_STRING) &&
				(getExpression(1).getType().getTypeCode() ==
				PType.TYPECODE_STRING) &&
				(getExpression(2).getType().isa(PTypeSpace.INT))) {
				result = new BasicVariableOperand(type);
				verified = true;
			}
			else
			//it is not a boolean type
				throw new TypingException("Attempted argument types: String, String, int");
		}
		else
		//Operands not yet assigned
			throw new MalformedExpressionException("Null children value");
		return result;
	}

	public int evaluate(String inputString, String subString, int fromIndex) {
		return inputString.indexOf(subString, fromIndex - 1);
	}

	public String getOperatorString() {
		return Operator.INDEXEDLOCATE;
	}
}