/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.medor.expression.lib;

import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.jorm.type.api.PType;

/**
 * This unary operator permit to change the type of an expression to another.
 * It is a kind of cast operation.
 *
 * @author S.Chassande-Barrioz
 */
public class TypeConverter extends BasicUnaryOperator {

	/**
	 * Builds a TypeConverter.
	 *
	 * @param newType is the new PType of the expression. (cannot be null).
	 */
	public TypeConverter(PType newType) throws MalformedExpressionException {
		super();
		if (newType == null) {
			throw new  MalformedExpressionException("Null PType not supported");
		}
		type = newType;
	}

	/**
	 * Builds a TypeConverter.
	 *
	 * @param operand is the sub expression which the type must be converted.
	 * @param newType is the new PType of the expression. (cannot be null).
	 */
	public TypeConverter(Expression operand,
						 PType newType) throws MalformedExpressionException {
		this(newType);
		setExpression(0, operand);
	}

	public String getOperatorString() {
		return "cast(" + type.getJavaName() + ")";
	}

	public Operand evaluate(ParameterOperand[] values, Object o)
		throws ExpressionException {
		if (expressions[0] == null) {
			throw new ExpressionException("No inner operand specified");
		}
		Operand subOperand = expressions[0].evaluate(values, o);
		PType subType = expressions[0].getType();

		switch(type.getTypeCode()) {
		case PType.TYPECODE_BOOLEAN:
		case PType.TYPECODE_OBJBOOLEAN:
			{
				boolean res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = subOperand.getBoolean();
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte() > 0;
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort() > 0;
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = subOperand.getInt() > 0;
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = subOperand.getLong() > 0;
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = subOperand.getFloat() > 0;
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = subOperand.getDouble() > 0;
					break;
				case PType.TYPECODE_STRING:
					res = Boolean.valueOf(subOperand.getString()).booleanValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_BOOLEAN) {
					result.setValue(res);
				} else {
					result.setValue(res ? Boolean.TRUE : Boolean.FALSE);
				}
			}

		case PType.TYPECODE_CHAR:
		case PType.TYPECODE_OBJCHAR:
			{
				char res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_CHAR:
				case PType.TYPECODE_OBJCHAR:
					res = subOperand.getChar();
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = (char) subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = (char) subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = (char) subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = (char) subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = (char) subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (char) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					{
						String str = subOperand.getString();
						if (str == null) {
							res = (char) 0;
						} else {
							res = str.charAt(0);
						}
						break;
					}
				default :
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_CHAR) {
					result.setValue(res);
				} else {
					result.setValue(new Character(res));
				}
			}

		case PType.TYPECODE_BYTE:
		case PType.TYPECODE_OBJBYTE:
			{
				byte res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (byte) (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = (byte) subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = (byte) subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = (byte) subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = (byte) subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (byte) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Byte.valueOf(subOperand.getString()).byteValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_BYTE) {
					result.setValue(res);
				} else {
					result.setValue(new Byte(res));
				}
			}

		case PType.TYPECODE_SHORT:
		case PType.TYPECODE_OBJSHORT:
			{
				short res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (short) (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = (short) subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = (short) subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = (short) subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (short) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Short.valueOf(subOperand.getString()).shortValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_SHORT) {
					result.setValue(res);
				} else {
					result.setValue(new Short(res));
				}
			}

		case PType.TYPECODE_INT:
		case PType.TYPECODE_OBJINT:
			{
				int res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = (int) subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = (int) subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (int) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Integer.valueOf(subOperand.getString()).intValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_INT) {
					result.setValue(res);
				} else {
					result.setValue(new Integer(res));
				}
			}

		case PType.TYPECODE_LONG:
		case PType.TYPECODE_OBJLONG:
			{
				long res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = (long) subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (long) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Long.valueOf(subOperand.getString()).longValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_LONG) {
					result.setValue(res);
				} else {
					result.setValue(new Long(res));
				}
			}

		case PType.TYPECODE_FLOAT:
		case PType.TYPECODE_OBJFLOAT:
			{
				float res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = (float) subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Float.valueOf(subOperand.getString()).floatValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_LONG) {
					result.setValue(res);
				} else {
					result.setValue(new Float(res));
				}
			}

		case PType.TYPECODE_DOUBLE:
		case PType.TYPECODE_OBJDOUBLE:
			{
				double res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = (subOperand.getBoolean() ? 1 : 0);
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = Double.valueOf(subOperand.getString()).doubleValue();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				if (type.getTypeCode() == PType.TYPECODE_LONG) {
					result.setValue(res);
				} else {
					result.setValue(new Float(res));
				}
			}
			break;

		case PType.TYPECODE_STRING:
			{
				String res;
				switch (subType.getTypeCode()) {
				case PType.TYPECODE_BOOLEAN:
				case PType.TYPECODE_OBJBOOLEAN:
					res = "" + subOperand.getBoolean();
					break;
				case PType.TYPECODE_BYTE:
				case PType.TYPECODE_OBJBYTE:
					res = "" + subOperand.getByte();
					break;
				case PType.TYPECODE_SHORT:
				case PType.TYPECODE_OBJSHORT:
					res = "" + subOperand.getShort();
					break;
				case PType.TYPECODE_INT:
				case PType.TYPECODE_OBJINT:
					res = "" + subOperand.getInt();
					break;
				case PType.TYPECODE_LONG:
				case PType.TYPECODE_OBJLONG:
					res = "" + subOperand.getLong();
					break;
				case PType.TYPECODE_FLOAT:
				case PType.TYPECODE_OBJFLOAT:
					res = "" + subOperand.getFloat();
					break;
				case PType.TYPECODE_DOUBLE:
				case PType.TYPECODE_OBJDOUBLE:
					res = "" + subOperand.getDouble();
					break;
				case PType.TYPECODE_STRING:
					res = subOperand.getString();
					break;
				default:
					throw new ExpressionException("Impossible to convert "
						+ subType.getJavaName() + " to " + type.getJavaName());
				}
				result.setValue(res);
			}
			break;

		//Unsupported type
		case PType.TYPECODE_SERIALIZED:
		case PType.TYPECODE_DATE:
		case PType.TYPECODE_REFERENCE:
		case PType.TYPECODE_BIGDECIMAL:
		case PType.TYPECODE_BIGINTEGER:
		case PType.TYPECODE_BYTEARRAY:
		case PType.TYPECODE_CHARARRAY:
		default:
			throw new ExpressionException("Impossible to convert "
				+ subType.getJavaName() + " to " + type.getJavaName());
		}
		return result;
	}

	public Operand compileExpression()
		throws ExpressionException, MalformedExpressionException {
		if (result == null) {
			result = new BasicVariableOperand(type);
		}
		expressions[0].compileExpression();
		return result;
	}
}
